#!/usr/bin/perl -w

# Fishkin’s Yggdrasil nodes pinger 0.0.5.
# © Gamliel Fishkin, 2023-2024. CC BY-SA 4.0.
# https://fynpinger.fishk.in/

use strict;
use warnings;

print 'Fishkin’s Yggdrasil nodes pinger 0.0.5.' . "\n" . '© Gamliel Fishkin, 2023-2024. CC BY-SA 4.0.' . "\n" . 'https://fynpinger.fishk.in/' . "\n\n";
my %conf;
if ($ENV{LANG}) {
  $conf{lng} = (split(/_/, (split(/\./, $ENV{LANG}))[0]))[0];
}
else {
  $conf{lng} = 'en';
}

{
  my @warn;
  if ($conf{lng} eq 'ru') {
    @warn = ('Ваша операционная система не поддерживается. ', 'Её поддержка возможна в будущем, но сейчас используйте GNU/Linux.', 'Используйте GNU/Linux.');
  }
  elsif ($conf{lng} eq 'eo') {
    @warn = ('Via operacia sistemo ne estas subtenata. ', 'Ĝia subteno eblas en estonteco, sed nun bonvolu uzi GNU/Linukson.', 'Bonvolu uzi GNU/Linukson.');
  }
  else {
    @warn = ('Your operating system is not supported. ', 'Its support is possible in future, but for now please use GNU/Linux.', 'Please use GNU/Linux.');
  }
  if (`which uname 2> /dev/null`) {
    my $kernel = `uname -s 2> /dev/null`;
    chomp($kernel);
    if ($kernel ne 'Linux') {
      if ($kernel =~ m/.*BSD$/) {
        die $warn[0] . $warn[1] . "\n";
      }
      else {
        die $warn[0] . $warn[2] . "\n";
      }
    }
  }
  else {
    die $warn[0] . $warn[2] . "\n";
  }
}

%conf = (
  %conf,
  ip => 'both',
  proto => 'quic,tls',
  hostnames => 'nocf',
  offline => 'none',
  max => 42,
  blocklist => 'no',
  morehosts => 'no',
  to => 'both',
);
my $dir;
{
  my @dir = split(/\//, $ENV{_});
  my @pwd;
  if ($dir[0] ne '') {
    @pwd = split(/\//, $ENV{PWD});
    while ($dir[0] eq '..' || $dir[0] eq '.') {
      pop @pwd if $dir[0] eq '..';
      shift @dir;
    }
  }
  $dir[-1] = '';
  $dir = join '/', @pwd, @dir;
}

my (%lng);
{
  my ($errcf, $errcl);
  my $errors = '';
  if (-r $dir . 'fynpinger.conf') {
    open(CONF, '< ' . $dir . 'fynpinger.conf');
    my @conf = <CONF>;
    close(CONF);
    chomp(@conf);
    $errcf = &conf('file', @conf);
  }
  elsif (-e $dir . 'fynpinger.conf') {
    $errors = 'unreadable';
  }
  $errcl = &conf('argv', @ARGV);
  $conf{lng} = 'en' if $conf{lng} !~ m/^(en|eo|ru)$/;
  if ($conf{lng} eq 'ru') {
    %lng = (
      A5 => 'Конфигурационный файл существует, но не читается. Проверьте разрешения.',
      B5 => 'Ошибка(-и) в конфигурационном файле',
      C5 => 'Ошибка(-и) в командной строке',
      D5 => 'Подготовка. Пожалуйста, подождите',
      E5 => 'Вы хотели проверить IPv4-адреса, но здесь нет IPv4.',
      F5 => 'Вы хотели проверить IPv6-адреса, но здесь нет IPv6.',
      G5 => 'Здесь нет IPv6; только IPv4-адреса будут проверены.',
      H5 => 'Здесь нет IPv4; только IPv6-адреса будут проверены.',
      I5 => 'Сеть недоступна: нет ни IPv4, ни IPv6.',
      J5 => 'Каталог «',
      K5 => '» недоступен на запись. Проверьте разрешения.',
      L5 => 'Файл «',
      M5 => '» недоступен на запись. Проверьте разрешения.',
      N5 => 'Пожалуйста, установите пакет wget или curl.',
      O5 => 'Пожалуйста, установите пакет bind9-dnsutils или knot-dnsutils. Сейчас будут проверены только IP-адреса.',
      P5 => 'Подготовка закончена. Пингую; пожалуйста, будьте терпеливы и подождите приблизительно от',
      Q5 => 'до',
      R5 => 'секунд (или чуть больше)',
      S5 => 'секунды (или чуть больше)',
      T5 => 'сек. Пинговано',
      U5 => 'найдено',
      V5 => 'без ответа',
      W5 => 'Не  ответили',
      X5 => 'Быстрые узлы',
      Y5 => 'Результат сохранён в файле «',
      Z5 => '».',
    );
  }
  elsif ($conf{lng} eq 'eo') {
    %lng = (
      A5 => 'La konfugura dosiero ekzistas, sed ne estas legebla. Kontrolu permesojn.',
      B5 => 'Eraro(j) en la konfigura dosiero',
      C5 => 'Eraro(j) en la komanda linio',
      D5 => 'Preparado. Bonvolu atendi',
      E5 => 'Vi deziris testi IPv4-adresojn, sed ĉi tie ne estas IPv4.',
      F5 => 'Vi deziris testi IPv6-adresojn, sed ĉi tie ne estas IPv6.',
      G5 => 'Ĉi tie ne estas IPv6; nur IPv4-adresoj estos testitaj.',
      H5 => 'Ĉi tie ne estas IPv4; nur IPv6-adresoj estos testitaj.',
      I5 => 'Reto estas neatingebla: estas nek IPv4 nek IPv6.',
      J5 => 'La katalogo «',
      K5 => '» ne estas skribebla. Kontrolu permesojn.',
      L5 => 'La dosiero «',
      M5 => '» ne estas skribebla. Kontrolu permesojn.',
      N5 => 'Bonvolu instali pakaĵon wget aŭ curl.',
      O5 => 'Bonvolu instali pakaĵon bind9-dnsutils aŭ knot-dnsutils. Nun nur IP-adresoj estos testitaj.',
      P5 => 'Preparoj estas finitaj. Tintado (ping); bonvolu esti pacienca kaj atendi proksimume inter',
      Q5 => 'kaj',
      R5 => 'sekundojn (aŭ iomete pli)',
      S5 => 'sekundojn (aŭ iomete pli)',
      T5 => 'sek. Tintita(j)',
      U5 => 'trovita(j)',
      V5 => 'sen respondo',
      W5 => 'Ne  respondis',
      X5 => 'Rapidaj nodoj',
      Y5 => 'La rezulto estas konservita en la dosiero «',
      Z5 => '».',
    );
  }
  else {
    %lng = (
      A5 => 'The configuration file exists, but is not readable. Check permissions.',
      B5 => 'Error(s) in the configuration file',
      C5 => 'Error(s) on the command line',
      D5 => 'Preparing. Please wait',
      E5 => 'You wanted to test IPv4 addresses, but there is no IPv4.',
      F5 => 'You wanted to test IPv6 addresses, but there is no IPv6.',
      G5 => 'There is no IPv6; only IPv4 addresses will be tested.',
      H5 => 'There is no IPv4; only IPv6 addresses will be tested.',
      I5 => 'Network is unreachable: no IPv4 neither IPv6.',
      J5 => 'The directory ‘',
      K5 => '’ is not writable. Check permissions.',
      L5 => 'The file ‘',
      M5 => '’ is not writable. Check permissions.',
      N5 => 'Please install package wget or curl.',
      O5 => 'Please install package bind9-dnsutils or knot-dnsutils. Only IP addresses will be tested now.',
      P5 => 'Preparations are finished. Pinging; please be patient and wait approximately between',
      Q5 => 'and',
      R5 => 'seconds (or a bit more)',
      S5 => 'seconds (or a bit more)',
      T5 => 'sec. Pinged',
      U5 => 'found',
      V5 => 'no reply',
      W5 => 'Not replied',
      X5 => 'Quick nodes',
      Y5 => 'The result is saved in the file ‘',
      Z5 => '’.',
    );
  }
  if ($errcf || $errcl || $errors) {
    if ($errors eq 'unreadable') {
      $errors = $lng{A5} . "\n";
    }
    elsif ($errcf) {
      $errors = $lng{B5} . ': ' . $errcf . ".\n";
    }
    if ($errcl) {
      $errors .= $lng{C5} . ': ' . $errcl . ".\n";
    }
    die $errors . "\n" if $errors;
  }
}

print $lng{D5} . '...';
{ # rs here is an abbreviation for root-servers; G does not reply to ping
  my @rs4 = qw(198.41.0.4 170.247.170.2 192.33.4.12 199.7.91.13 192.203.230.10 192.5.5.241 198.97.190.53 192.36.148.17 192.58.128.30 193.0.14.129 199.7.83.42 202.12.27.33);
  my @rs6 = qw(2001:503:ba3e::2:30 2801:1b8:10::b 2001:500:2::c 2001:500:2d::d 2001:500:a8::e 2001:500:2f::f 2001:500:1::53 2001:7fe::53 2001:503:c27::2:30 2001:7fd::1 2001:500:9f::42 2001:dc3::35);
  my $niu = '(Network is unreachable|non-recoverable failure in name resolution|Cannot resolve)';
  my $rs = int(rand(12));
  if ($conf{ip} eq 'v4') {
    my @p4 = `LANG=C ping -qw2 $rs4[$rs] 2>&1`;
    if ($p4[0] =~ m/$niu/) {
      die "\r" . $lng{E5} . "\n";
    }
    elsif ($p4[-2] =~ m/100% packet loss/) {
      $rs = int(rand(12));
      if ((`LANG=C ping -qw2 $rs4[$rs]`)[-2] =~ m/100% packet loss/) {
        $rs = int(rand(12));
        die "\r" . $lng{E5} . "\n" if (`LANG=C ping -qw2 $rs4[$rs]`)[-2] =~ m/100% packet loss/;
      }
    }
  }
  elsif ($conf{ip} eq 'v6') {
    my @p6 = `LANG=C ping -qw2 $rs6[$rs] 2>&1`;
    if ($p6[0] =~ m/$niu/) {
      die "\r" . $lng{F5} . "\n";
    }
    elsif ($p6[-2] =~ m/100% packet loss/) {
      $rs = int(rand(12));
      if ((`LANG=C ping -qw2 $rs6[$rs]`)[-2] =~ m/100% packet loss/) {
        $rs = int(rand(12));
        die "\r" . $lng{F5} . "\n" if (`LANG=C ping -qw2 $rs6[$rs]`)[-2] =~ m/100% packet loss/;
      }
    }
  }
  else {
    my @p4 = `LANG=C ping -qw2 $rs4[$rs] 2>&1`;
    my @p6 = `LANG=C ping -qw2 $rs6[$rs] 2>&1`;
    if ($p4[0] !~ m/$niu/ && $p4[-2] !~ m/100% packet loss/ && ($p6[0] =~ m/$niu/ || $p6[-2] =~ m/100% packet loss/)) {
      if ($p6[0] =~ m/$niu/) {
        print "\r" . $lng{G5} . "\n";
        $conf{ip} = 'v4';
      }
      else {
        $rs = int(rand(12));
        @p4 = `LANG=C ping -qw2 $rs4[$rs] 2>&1`;
        @p6 = `LANG=C ping -qw2 $rs6[$rs] 2>&1`;
        if ($p4[0] !~ m/$niu/ && $p4[-2] !~ m/100% packet loss/ && $p6[-2] =~ m/100% packet loss/) {
          $rs = int(rand(12));
          @p4 = `LANG=C ping -qw2 $rs4[$rs] 2>&1`;
          @p6 = `LANG=C ping -qw2 $rs6[$rs] 2>&1`;
          if ($p4[0] !~ m/$niu/ && $p4[-2] !~ m/100% packet loss/ && $p6[-2] =~ m/100% packet loss/) {
            print "\r" . $lng{G5} . "\n";
            $conf{ip} = 'v4';
          }
        }
      }
    }
    elsif ($p6[0] !~ m/$niu/ && $p6[-2] !~ m/100% packet loss/ && ($p4[0] =~ m/$niu/ || $p4[-2] =~ m/100% packet loss/)) {
      if ($p4[0] =~ m/$niu/) {
        print "\r" . $lng{H5} . "\n";
        $conf{ip} = 'v6';
      }
      else {
        $rs = int(rand(12));
        @p4 = `LANG=C ping -qw2 $rs4[$rs] 2>&1`;
        @p6 = `LANG=C ping -qw2 $rs6[$rs] 2>&1`;
        if ($p6[0] !~ m/$niu/ && $p6[-2] !~ m/100% packet loss/ && $p4[-2] =~ m/100% packet loss/) {
          $rs = int(rand(12));
          @p4 = `LANG=C ping -qw2 $rs4[$rs] 2>&1`;
          @p6 = `LANG=C ping -qw2 $rs6[$rs] 2>&1`;
          if ($p6[0] !~ m/$niu/ && $p6[-2] !~ m/100% packet loss/ && $p4[-2] =~ m/100% packet loss/) {
            print "\r" . $lng{H5} . "\n";
            $conf{ip} = 'v6';
          }
        }
      }
    }
    elsif ($p4[0] =~ m/$niu/ && $p6[0] =~ m/$niu/) {
      print "\r" . $lng{I5} . "\n";
      exit;
    }
    elsif ($p4[-2] =~ m/100% packet loss/ && $p6[-2] =~ m/100% packet loss/) {
      $rs = int(rand(12));
      if ((`LANG=C ping -qw2 $rs4[$rs]`)[-2] =~ m/100% packet loss/ && (`LANG=C ping -qw2 $rs6[$rs]`)[-2] =~ m/100% packet loss/) {
        $rs = int(rand(12));
        if ((`LANG=C ping -qw2 $rs4[$rs]`)[-2] =~ m/100% packet loss/ && (`LANG=C ping -qw2 $rs6[$rs]`)[-2] =~ m/100% packet loss/) {
          print "\r" . $lng{I5} . "\n";
          exit;
        }
      }
    }
  }
}

my ($i, @host, @port);
my @list = `yggdrasilctl getpeers 2> /dev/null | grep Out | cut -d' ' -f1 | cut -f1`;
my @proto = split(/,/, $conf{proto});
{
  my $fn = 'nodes.html';
  if (!-w $dir) {
    die $lng{J5} . $dir . $lng{K5} . "\n";
  }
  elsif (-e $dir . $fn && !-w $dir . $fn) {
    die $lng{L5} . $dir . $fn . $lng{M5} . "\n";
  }
  else {
    my $clurl = ' https://publicpeers.neilalexander.dev/';
    my $url;
    if (@list) {
      chomp(@list);
      $url = ' http://[200:2688:699a:ce30:5897:cd3d:f999:8782]/';
    }
    else {
      $url = $clurl;
    }
    my $cmd;
    if (`which wget 2> /dev/null`) {
      $cmd = 'wget -t2 -T10 -w10 -Nq --default-page=' . $fn;
      system('cd ' . $dir . '; ' . $cmd . $url . ' || ' . $cmd . $clurl);
    }
    elsif (`which curl 2> /dev/null`) {
      $cmd = 'curl -m10 --retry 1 --retry-delay 10 -Rso ' . $fn;
      if (-e $dir . $fn) {
        system('cd ' . $dir . '; ' . $cmd . ' -z ' . $fn . $url . ' || ' . $cmd . ' -z ' . $fn . $clurl);
      }
      else {
        system('cd ' . $dir . '; ' . $cmd . $url . ' || ' . $cmd . $clurl);
      }
    }
    else {
      die $lng{N5} . "\n";
    }
  }
  my @host4 = ('');
  my @port4 = ('');
  my @host6 = ('');
  my @port6 = ('');
  my (@ygg, $proto);
  {
    my $qproto;
    if (0 == $#proto) {
      $proto = $proto[0];
      $qproto = $proto[0];
    }
    else {
      $proto = '(' . (join '|', @proto) . ')';
      $qproto = '(?:' . (join '|', @proto) . ')';
    }
    if (-e $dir . $fn) {
      open(YGG, '< ' . $dir . $fn);
      @ygg = <YGG>;
      close(YGG);
      @ygg = grep(/$qproto:\/\//, @ygg);
      foreach (@ygg) {
        m/^\<td.*\>(quic|tcp|tls)(:\/\/[a-z0-9\.\[:\]]+?:[0-9]+)(\?key=[0-9a-f]+|)\<\/td\>.*\<td.*\>(online.+?|offline.+?)\<\/td\>.*$/;
        $_ = $1 . $2 . $3 . ';' . $4;
      }
    }
    if ($conf{morehosts} ne 'no') {
      my @morehosts = split(/,/, $conf{morehosts});
      foreach (@morehosts) {
        if (m/^$qproto:\/\/(?:[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+:[0-9]+.*|\[[23][0-9a-f]{3}:[0-9a-f:]+\]:[0-9]+.*|[a-z0-9\.-]+\.[a-z]{2,6}:[0-9]+.*)$/) {
          push @ygg, $_ . ';unknown';
        }
      }
    }
  }
  my $dig;
  if (`which kdig 2> /dev/null`) {
    $dig = 'kdig';
  }
  elsif (`which dig 2> /dev/null`) {
    $dig = 'dig';
  }
  else {
    print "\r" . $lng{O5} . "\n";
    $conf{hostnames} = 'none';
  }
  my ($hostport, $host, $port, $uri, %dig4, %dig6);
  mark10: foreach (@ygg) {
    {
      my ($addr, $status) = split(/;/);
      if ($conf{offline} ne 'all' && $status =~ m/^offline/) {
        if ($conf{offline} eq 'none' || ($conf{offline} eq 'day' && $status =~ m/week/) || ($conf{offline} eq 'hour' && $status =~ m/(day|week)/)) {
          next mark10;
        }
      }
      ($proto, $hostport) = split(/:\/\//, $addr);
    }
    $proto = 'zcp' if $proto eq 'tcp'; # to place 'tcp' after 'tls' at sorting
    my (@blocklist, $imax);
    @blocklist = split(/,/, $conf{blocklist}) if $conf{blocklist} ne 'no';
    if ($hostport =~ m/^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}:[0-9]{1,5}.*/) {
      if ($conf{ip} =~ m/^(both|v4)$/) {
        ($host, $port) = split(/:/, $hostport);
        if (!grep(/^$host$/, @blocklist)) {
          $uri = $proto . '://addr:' . $port;
          if (!grep(/^$host$/, @host4)) {
            push @host4, $host;
            push @port4, $uri;
          }
          else {
            $imax = $#host4;
            mark20: for $i (0 .. $imax) {
              if ($host eq $host4[$i]) {
                if ($uri ne $port4[$i] && !grep(/^$uri$/, (split(/,/, $port4[$i])))) {
                  $port4[$i] .= ',' . $uri;
                }
                last mark20;
              }
            }
          }
        }
      }
    }
    elsif ($hostport =~ m/^\[([0-9a-f:]{2,39}\]:[0-9]{1,5}.*)/) {
      $hostport = $1;
      if ($conf{ip} =~ m/^(both|v6)$/) {
        ($host, $port) = split(/]:/, $hostport);
        if (!grep(/^$host$/, @blocklist)) {
          $uri = $proto . '://addr:' . $port;
          if (!grep(/^$host$/, @host6)) {
            push @host6, $host;
            push @port6, $uri;
          }
          else {
            $imax = $#host6;
            mark30: for $i (0 .. $imax) {
              if ($host eq $host6[$i]) {
                if ($uri ne $port6[$i] && !grep(/^$uri$/, (split(/,/, $port6[$i])))) {
                  $port6[$i] .= ',' . $uri;
                }
                last mark30;
              }
            }
          }
        }
      }
    }
    elsif ($conf{hostnames} =~ m/^(all|nocf)$/) {
      ($host, $port) = split(/:/, $hostport);
      if (!grep(/^$host$/, @blocklist)) {
        my (@dig, $hostn);
        if ($conf{ip} =~ m/^(both|v4)$/) {
          if ($dig4{$host} && $dig4{$host} eq 'none') {
            next mark10;
          }
          elsif ($dig4{$host}) {
            @dig = split(/\n/, $dig4{$host});
          }
          elsif (@dig = `$dig +short +nocookie +timeout=1 +retry=1 $host a 2>/dev/null`) {
            $dig4{$host} = join '', @dig;
          }
          else {
            $dig4{$host} = 'none';
            next mark10;
          }
          if ($conf{hostnames} eq 'all' || $dig[0] !~ m/\.pacloudflare\.com\.$/) {
            $hostn = $dig[-1];
            chomp($hostn);
            if ($hostn !~ m/[a-z]/ && !grep(/^$hostn$/, @blocklist)) {
              $uri = $proto . '://addr:' . $port;
              if (!grep(/^$hostn$/, @host4)) {
                push @host4, $hostn;
                push @port4, $uri;
              }
              else {
                $imax = $#host4;
                mark40: for $i (0 .. $imax) {
                  if ($hostn eq $host4[$i]) {
                    if ($uri ne $port4[$i] && !grep(/^$uri$/, (split(/,/, $port4[$i])))) {
                      $port4[$i] .= ',' . $uri;
                    }
                    last mark40;
                  }
                }
              }
            }
          }
        }
        if ($conf{ip} =~ m/^(both|v6)$/) {
          if ($dig6{$host} && $dig6{$host} eq 'none') {
            next mark10;
          }
          elsif ($dig6{$host}) {
            @dig = split(/\n/, $dig6{$host});
          }
          elsif (@dig = `$dig +short +nocookie +timeout=1 +retry=1 $host aaaa 2>/dev/null`) {
            $dig6{$host} = join '', @dig;
          }
          else {
            $dig6{$host} = 'none';
            next mark10;
          }
          if ($conf{hostnames} eq 'all' || $dig[0] !~ m/\.pacloudflare\.com\.$/) {
            $hostn = $dig[-1];
            chomp($hostn);
            if ($hostn =~ m/:/ && !grep(/^$hostn$/, @blocklist)) {
              $uri = $proto . '://addr:' . $port;
              if (!grep(/^$hostn$/, @host6)) {
                push @host6, $hostn;
                push @port6, $uri;
              }
              else {
                $imax = $#host6;
                mark50: for $i (0 .. $imax) {
                  if ($hostn eq $host6[$i]) {
                    if ($uri ne $port6[$i] && !grep(/^$uri$/, (split(/,/, $port6[$i])))) {
                      $port6[$i] .= ',' . $uri;
                    }
                    last mark50;
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  shift @host4;
  shift @port4;
  shift @host6;
  shift @port6;
  @host = (@host4, @host6);
  @port = (@port4, @port6);
}

my (@lowpings, @zeroping);
{
  my $lenp = length($conf{max} - 1) + 4;
  my $imax = $#host + 1;
  my $lens = length(int($imax * 2.2 + 0.5));
  my $lenh = length($imax) . 'd';
  {
    my $min = int($imax * 1.9 + 0.5);
    my $max = int($imax * 2.1 + 0.5);
    if ($max % 10 != 1 || $max % 100 == 11) {
      $max .= ' ' . $lng{R5};
    }
    else {
      $max .= ' ' . $lng{S5};
    }
    print "\r" . $lng{P5} . ' ' . $min . ' ' . $lng{Q5} . ' ' . $max . ".\n";
  }
  my (@ping, $ping, $tnow, $n);
  my ($pinged, $found, $noreply) = (0, 0, 0);
  my $tstart = time();
  for ($i = 0; $i < $imax; ++$i) {
    @ping = `LANG=C ping -Aqw2 $host[$i]`;
    if ($ping[-2] !~ m/100% packet loss/) {
      $ping = (split(/\//, (split(/ /, $ping[-1]))[3]))[0];
      if ($ping < $conf{max}) {
        $ping = sprintf("%$lenp.3f", $ping);
        if ($port[$i] !~ m/,/) {
          my $addr = $port[$i];
          if ($host[$i] =~ m/:/) {
            $n = $addr;
            $addr =~ s/addr/\[$host[$i]\]/;
            $n =~ s/addr/\\\[$host[$i]\\\]/;
          }
          else {
            $addr =~ s/addr/$host[$i]/;
            $n = $addr;
          }
          push @lowpings, $ping . '  ' . $addr . 'n' . &n($n, ' ');
        }
        else {
          my @addr = split(/,/, $port[$i]);
          if ($host[$i] =~ m/:/) {
            foreach (@addr) {
              $n = $_;
              s/addr/\[$host[$i]\]/;
              $n =~ s/addr/\\\[$host[$i]\\\]/;
              push @lowpings, $ping . '  ' . $_ . 'n' . &n($n, ' ');
            }
          }
          else {
            foreach (@addr) {
              s/addr/$host[$i]/;
              push @lowpings, $ping . '  ' . $_ . 'n' . &n($_, ' ');
            }
          }
        }
        ++$found;
      }
    }
    else {
      if ($port[$i] !~ m/,/) {
        my $addr = $port[$i];
        if ($host[$i] =~ m/:/) {
          $n = $addr;
          $addr =~ s/addr/\[$host[$i]\]/;
          $n =~ s/addr/\\\[$host[$i]\\\]/;
        }
        else {
          $addr =~ s/addr/$host[$i]/;
          $n = $addr;
        }
        push @zeroping, &n($n, '') . $addr;
      }
      else {
        my @addr = split(/,/, $port[$i]);
        my @zp;
        foreach (@addr) {
          if ($host[$i] =~ m/:/) {
            $n = $_;
            s/addr/\[$host[$i]\]/;
            $n =~ s/addr/\\\[$host[$i]\\\]/;
          }
          else {
            s/addr/$host[$i]/;
            $n = $_;
          }
          push @zp, &n($n, '') . $_;
        }
        push @zeroping, (join ', ', @zp);
      }
      ++$noreply;
    }
    ++$pinged;
    $tnow = time() - $tstart;
    my $line = sprintf("%${lens}d", $tnow) . ' ' . $lng{T5} . ': ' . sprintf("%$lenh", $pinged) . ', ' . $lng{U5};
    $line .= ': ' . sprintf("%$lenh", $found) . '; ' . $lng{V5} . ': ' . sprintf("%$lenh", $noreply) . '...';
    print "\r" . $line;
  }
}
print "\b\b  \n";
@host = ();
@port = ();

{
  @lowpings = sort @lowpings;
  my @n;
  if (grep(/^tcp$/, @proto)) {
    s/zcp/tcp/g foreach (@zeroping);
    foreach (@lowpings) {
      s/zcp/ tcp/;
      s/tls/ tls/;
      @n = split(/n/);
      $_ = $n[1] . ' ' . $n[0];
    }
  }
  else {
    foreach (@lowpings) {
      s/tls/ tls/;
      @n = split(/n/);
      $_ = $n[1] . ' ' . $n[0];
    }
  }
}
my $result = "\n";
$result .= join "\n", ('======== ' . $lng{W5} . ' ========', @zeroping, '') if $#zeroping > -1;
$result .= join "\n", ('======== ' . $lng{X5} . ' ========', @lowpings, '') if $#lowpings > -1;
my $msg = '';
if ($conf{to} =~ m/^(?:file|both)$/) {
  opendir(DIR, $dir);
  my $fn = 'fynpinger-result' . (grep(/^fynpinger-re[sz]ult[0-9]+\.lst$/, readdir DIR)) . '.lst'; # previous versions used the wrong spelling "rezult"
  closedir(DIR);
  my %lngname = (en => 'English', eo => 'Esperanto', ru => 'Русский');
  open(RESULT, '> ' . $dir . $fn);
  print RESULT '# ' . $lngname{$conf{lng}} . $result;
  close(RESULT);
  $msg = "\n" . $lng{Y5} . $dir . $fn . $lng{Z5} . "\n";
}
print $result if $conf{to} =~ m/^(?:screen|both)$/;
print $msg;

sub conf {
  my ($conf, @conf) = (@_);
  my @err;
  foreach (@conf) {
    if ($conf eq 'file' && (m/^\#/ || m/^$/)) {}
    elsif (m/^lng=(en|eo|ru)$/) {
      $conf{lng} = $1;
    }
    elsif (m/^lng=(?:auto|[a-z]{2,3})$/) {
      if (defined $ENV{LANG}) {
        $conf{lng} = (split(/_/, (split(/\./, $ENV{LANG}))[0]))[0];
      }
      else {
        $conf{lng} = 'en';
      }
    }
    elsif (m/^ip=(both|v4|v6)$/) {
      $conf{ip} = $1;
    }
    elsif (m/^proto=(.+)$/) {
      if ($1 eq 'all' || $1 eq 'both') { # 'both' is deprecated and in future versions can be removed
        $conf{proto} = 'quic,tcp,tls';
      }
      else {
        my $p = join ',', (sort split(/,/, $1));
        if ($p =~ m/^(?:quic,tcp,tls|quic,tcp|quic,tls|tcp,tls|quic|tcp|tls)$/) {
          $conf{proto} = $p;
        }
        else {
          push @err, 'proto=' . $p;
        }
      }
    }
    elsif (m/^hostnames=(all|nocf|none)$/) {
      $conf{hostnames} = $1;
    }
    elsif (m/^offline=(none|hour|day|all)$/) {
      $conf{offline} = $1;
    }
    elsif (m/^max=([1-9][0-9]*)$/) {
      $conf{max} = $1;
    }
    elsif (m/^blocklist=(.*)$/) {
      my $blocklist = $1;
      if ($blocklist eq 'no' || $blocklist eq '') {
        $conf{blocklist} = 'no';
      }
      elsif (defined $conf{blocklist} && $blocklist =~ m/^\+(.+)$/) {
        $conf{blocklist} .= ',' . $1;
      }
      else {
        $conf{blocklist} = $blocklist;
      }
    }
    elsif (m/^morehosts=(.*)$/) {
      my $morehosts = $1;
      if ($morehosts eq 'no' || $morehosts eq '') {
        $conf{morehosts} = 'no';
      }
      elsif (defined $conf{morehosts} && $morehosts =~ m/^\+(.+)$/) {
        $conf{morehosts} .= ',' . $1;
      }
      else {
        $conf{morehosts} = $morehosts;
      }
    }
    elsif (m/^to=(screen|file|both)$/) {
      $conf{to} = $1;
    }
    else {
      push @err, $_;
    }
  }
  my $err = join ', ', @err;
  return $err;
}

sub n {
  if (grep(/^$_[0]$/, @list)) {
    return '*';
  }
  else {
    my ($n0, $n1) = split(/:\/\//, $_[0], 2);
    if ($n0 eq 'tcp') {
      $n0 = 'socks';
    }
    elsif ($n0 eq 'tls') {
      $n0 = 'sockstls';
    }
    if (grep(/^$n0:\/\/(?:[a-z0-9\.]+\.[a-z]{2,6}|[0-9\.]{7,15}|\[[0-9a-f:]{2,39}\]):[0-9]+\/$n1$/, @list)) {
      return '%';
    }
    else {
      return $_[1];
    }
  }
}
